import hashlib
import json
from pathlib import Path
from typing import Any, Callable, Literal

import flask
import numpy as np
import requests

from . import arrays as array_utils
from . import web as web_utils


def main_page_template() -> flask.Response:
    return flask.make_response("Server is running.")


def respond_supervised_anonymizer_training_data(
    data: str | Path | tuple[np.ndarray, np.ndarray]
) -> flask.Response:
    """Alias function for `supervised_anonymizer_training_data_template`.

    Use this alias.
    """
    return supervised_anonymizer_training_data_template(data)


def supervised_anonymizer_training_data_template(
    data: str | Path | tuple[np.ndarray, np.ndarray]
) -> flask.Response:
    """Provide training data for supervised anonymizers in terms of a flask
    `Response` object.

    Parameters
    ----------
    data : str | Path | tuple[np.ndarray, np.ndarray]
        Either a path to an `.npz` file, containing two arrays with names
        "x_train" and "y_train"; or a tuple of two numpy arrays, namely
        `(x_train, y_train)`

    Returns
    -------
    flask.Response
        Two compressed NumPy arrays embedded in a flask response. Their names
        are "x_train" and "y_train". Their shape and types are
        challenge-specific.
    """
    if isinstance(data, (str, Path)):
        return _respond_array_from_disk(
            data, 500, "Could not load anonymizer training data from disk."
        )
    else:
        assert isinstance(data, ((np.ndarray, np.ndarray),))
        (x, y) = data
        return array_utils.respond_numpy_arrays(
            download_name="train_supervised_anonymizer.npz",
            x_train=x,
            y_train=y,
        )


def _respond_array_from_disk(
    array: str | Path, error_code: int, error_msg: str
) -> flask.Response:
    array = Path(array)
    return web_utils.apply_or_abort(
        condition=array.exists(),
        error_code=error_code,
        error_msg=error_msg,
    )(array_utils.respond_npz)(array)


def respond_supervised_anonymizer_validation_data(
    data: str | Path | np.ndarray,
    anonymizer_hostname: str,
    anonymizer_secret: str,
    anomed_hostname: str,
    anomed_port: int,
) -> flask.Response:
    """Alias for `supervised_anonymizer_validation_data_template`.

    Use this alias."""
    return supervised_anonymizer_validation_data_template(
        data, anonymizer_hostname, anonymizer_secret, anomed_hostname, anomed_port
    )


def supervised_anonymizer_validation_data_template(
    data: str | Path | np.ndarray,
    anonymizer_hostname: str,
    anonymizer_secret: str,
    anomed_hostname: str,
    anomed_port: int,
) -> flask.Response:
    """Provide validation data for supervised anonymizer in terms of a flask
    `Response` object.

    Parameters
    ----------
    data : str | Path | np.ndarray
        Either a path to an `.npz` file, containing only (!) one array with name
        "x_val" (no labels); or a tuple of two numpy arrays, namely
        `(x_train, y_train)`
    anonymizer_hostname: str
        The identifying hostname of the data-requesting anonymizer.
    anonymizer_secret: str
        The secret obtained at registration of the anonymizer.
    anomed_hostname: str
        The hostname of the AnoMed platform's main server.
    anomed_port: int
        The port of the AnoMed platform's main server.

    Returns
    -------
    flask.Response
        A compressed NumPy array embedded in a flask response. Its name is
        "x_val". Its shape and type is challenge-specific.
    """
    if not _authenticate_hostname(
        anonymizer_hostname, anonymizer_secret, anomed_hostname, anomed_port
    ):
        return _authentication_error(anonymizer_hostname, "Anonymizer")
    if isinstance(data, (str, Path)):
        return _respond_array_from_disk(
            data, 500, "Could not load anonymizer validation data from disk."
        )
    elif isinstance(data, np.ndarray):
        return array_utils.respond_numpy_arrays(
            download_name="validation_supervised_anonymizer.npz",
            x_val=data,
        )
    else:
        flask.abort(500, "Unsupported type for argument `data`.")


def credentials_from_request(request: flask.Request, key: str) -> Any:
    """Extract a single credential from the credentials header of a request.

    Parameters
    ----------
    request : flask.Request
        The request containing the credentials header.
    key : Iterable[str]
        Which single value to extract.

    Returns
    -------
    Any
        The value stored at given key.

    Raises
    ------
    KeyError
        If the credential header does not contain `key`.
    ValueError
        If the request does not contain a credential header, or if does but the
        JSON string is malformed.
    """
    if "credentials" in request.headers:
        try:
            credentials: dict[str, str] = json.loads(request.headers["credentials"])
            return credentials[key]
        except json.JSONDecodeError:
            raise ValueError("Credentials is not valid JSON.")
    else:
        raise ValueError("Headers do not contain credential information.")


def evaluate_supervised_estimator(
    ground_truth: np.ndarray,
    estimation_request: flask.Request,
    evaluate: Callable[[np.ndarray, np.ndarray], dict[str, float]],
    estimator_hostname: str,
    estimator_secret: str,
    anomed_hostname: str,
    anomed_port: int,
    challenge_hostname: str,
    challenge_secret: str,
) -> dict[str, float]:
    """Alias for evaluate_estimator_template.

    Use this alias."""
    return evaluate_estimator_template(
        ground_truth,
        estimation_request,
        evaluate,
        estimator_hostname,
        estimator_secret,
        anomed_hostname,
        anomed_port,
        challenge_hostname,
        challenge_secret,
    )


def evaluate_estimator_template(
    ground_truth: np.ndarray,
    estimation_request: flask.Request,
    evaluate: Callable[[np.ndarray, np.ndarray], dict[str, float]],
    estimator_hostname: str,
    estimator_secret: str,
    anomed_hostname: str,
    anomed_port: int,
    challenge_hostname: str,
    challenge_secret: str,
) -> dict[str, float]:
    """Evaluate an estimator using the evaluation function `evaluate` and, as a
    side effect, submit the results to the AnoMed platform.

    Parameters
    ----------
    ground_truth : np.ndarray
        What the estimator should have estimated.
    estimation_request : flask.Request
        The request containing what the estimator actually estimated.
    evaluate : Callable[[np.ndarray, np.ndarray], dict[str, float]]
        The challenge-specific utility metric function.
    estimator_hostname : str
        The identifying hostname of the estimator (anonymizer / PPML model).
    estimator_secret: str
        The secret obtained at registration of the estimator.
    anomed_hostname : str
        The hostname of the AnoMed platform.
    anomed_port : int
        The port of the AnoMed platform.
    challenge_hostname : str
        The challenge's hostname.
    challenge_secret : str
        The challenge's secret.

    Returns
    -------
    dict[str, float]
        The utility metrics gathered by `evaluate`.
    """
    if not _authenticate_hostname(
        estimator_hostname, estimator_secret, anomed_hostname, anomed_port
    ):
        return _authentication_error(estimator_hostname, "Estimator")
    if not _authorize_anonymizer(
        estimator_hostname, challenge_hostname, anomed_hostname, anomed_port
    ):
        return flask.abort(
            401,
            f"Estimator {estimator_hostname} not authorized for the challenge "
            f"{challenge_hostname}.",
        )
    estimation = array_utils.all_numpy_arrays_from_request(estimation_request)["pred"]
    evaluation = evaluate(estimation, ground_truth)
    evaluation_to_submit = dict(anonymizer_hostname=estimator_hostname, **evaluation)
    resp = _submit_single_evaluation(
        anomed_hostname,
        challenge_hostname,
        challenge_secret,
        anonymizer_evaluation=evaluation_to_submit,
    )
    resp.raise_for_status()
    return evaluation


def _authenticate_hostname(
    hostname: str,
    secret: str,
    anomed_hostname: str,
    anomed_port: int,
) -> bool:
    """Return `True` only if the hostname is authentic.

    That means return `True` only if the provided secret belongs to that
    hostname. Otherwise return `False`. Assuming the challenge/submission behind
    that hostname did not share its secret, we may be sure that the challenge /
    submission is actually what it claims to be. The AnoMed main server, which
    keeps record of the secrets, is asked for authentication.
    """
    resp = web_utils.json_to_remote(
        jsonable_payload=dict(submission_hostname=hostname, submission_secret=secret),
        host=anomed_hostname,
        route="/submissions/authenticate-submission",
        port=anomed_port,
        how="GET",
    )
    return resp.status_code == 200


def _authentication_error(
    hostname: str, hostname_label: str | None, error_code: int = 401
):
    if hostname_label is None:
        hostname_label = "Hostname"
    flask.abort(error_code, f"{hostname_label} {hostname} is not authentic.")


def _authorize_anonymizer(
    anonymizer_hostname: str,
    challenge_hostname: str,
    anomed_hostname: str,
    anomed_port: int,
) -> bool:
    """Return `True` only if the anonymizer actually belongs to the challenge."""
    resp = web_utils.json_to_remote(
        jsonable_payload=dict(
            anonymizer_hostname=anonymizer_hostname,
            challenge_hostname=challenge_hostname,
        ),
        host=anomed_hostname,
        route="/submissions/authorize-anonymizer",
        port=anomed_port,
        how="GET",
    )
    return resp.status_code == 200


def _authorize_deanonymizer(
    deanonymizer_hostname: str,
    anonymizer_hostname: str,
    anomed_hostname: str,
    anomed_port: int,
) -> bool:
    """Return `True` only if the deanonymizer actually belongs to the stated
    anonymizer.

    That means return `True` only if, during registration of the deanonymizer at
    the AneMed platform, the attack target identified by `anonymizer_hostname`
    was chosen.
    """
    resp = web_utils.json_to_remote(
        jsonable_payload=dict(
            deanonymizer_hostname=deanonymizer_hostname,
            anonymizer_hostname=anonymizer_hostname,
        ),
        host=anomed_hostname,
        route="/submissions/authorize-deanonymizer",
        port=anomed_port,
        how="GET",
    )
    return resp.status_code == 200


def _submit_single_evaluation(
    anomed_hostname: str,
    challenge_hostname: str,
    challenge_secret: str,
    anonymizer_evaluation: dict[str, str | float] | None = None,
    deanonymizer_evaluation: dict[str, str | float] | None = None,
) -> requests.Response:
    json: dict[str, Any] = dict(
        challenge_secret=challenge_secret,
    )
    if anonymizer_evaluation is not None:
        json["anonymizer_json"] = [anonymizer_evaluation]
    if deanonymizer_evaluation is not None:
        json["deanonymizer_json"] = [deanonymizer_evaluation]
    response = web_utils.json_to_remote(
        json,
        host=anomed_hostname,
        port=8000,
        route=f"/challenges/{challenge_hostname}/update-evaluation-view/",
    )
    return response


def respond_MIA_training_data(
    train: str | Path | tuple[np.ndarray, np.ndarray], seed: int | None
) -> flask.Response:
    """Alias for membership_inference_attack_training_data_template.

    Use this alias."""
    return membership_inference_attack_training_data_template(train, seed)


def membership_inference_attack_training_data_template(
    train: str | Path | tuple[np.ndarray, np.ndarray], seed: int | None
) -> flask.Response:
    """Respond numpy arrays ("x_train" and "y_train") for training a membership
    inference attack.

    Usually, membership inference attacks need both, data that has been used for
    training of the target PPML model (members) and data that has not been used
    for training of the target PPML model (non-members). This function picks
    random subsets of half the size of `x_train` and `y_train` and responds
    them, to provide "seen" data for training an MIA.

    Parameters
    ----------
    train : str | Path | tuple[np.ndarray, np.ndarray]
        Either a path to an `.npz` file, containing two arrays with names
        "x_train" and "y_train"; or a tuple of two numpy arrays, namely
        `(x_train, y_train)`. The array named "x_train" contains the features
        that have been seen by the attack target (members). The array named
        "y_train" contains labels/targets that have been seen by the attack
        victim (members).
    seed : int | None
        The randomness to use for creating the random subsets. If `None`, obtain
        fresh randomness from the environment.

    Returns
    -------
    flask.Response
        A flask response embedding two arrays, with name "x_train" and
        "y_train", having the same datatype and shape (except for the batch
        dimension) as the parameters `x_train` and `y_train`.
    """
    (x_train, y_train) = array_utils.load_numpy_array_from_uncertain_source_or_abort(
        train, ("x_train", "y_train")
    )
    return _membership_inference_attack_data(
        x_train, y_train, seed, "train_mia.npz", "x_train", "y_train"
    )


def _membership_inference_attack_data(x, y, seed, download_name, x_name, y_name):
    (mia_x, mia_y) = _membership_inference_data(x, y, seed)
    return array_utils.respond_numpy_arrays_dict(
        named_arrays={x_name: mia_x, y_name: mia_y},
        download_name=download_name,
    )


def _membership_inference_data(x, y, seed) -> tuple[np.ndarray, np.ndarray]:
    return _split_half(x, y, "left", seed)


def _membership_inference_complementary_data(
    x, y, seed
) -> tuple[np.ndarray, np.ndarray]:
    return _split_half(x, y, "right", seed)


def _split_half(
    array_1: np.ndarray,
    array_2: np.ndarray,
    which: Literal["left", "right"],
    seed: int | None,
) -> tuple[np.ndarray, np.ndarray]:
    assert len(array_1) == len(array_2)
    [(arr1_l, arr1_r), (arr2_l, arr2_r)] = array_utils.random_partitions(
        arrays=(array_1, array_2),
        total_length=len(array_1),
        desired_length=len(array_1) // 2,
        seed=seed,
    )
    if which == "left":
        return (arr1_l, arr2_l)
    elif which == "right":
        return (arr1_r, arr2_r)
    else:
        raise ValueError(which)


def respond_MIA_validation_data(
    val: str | Path | tuple[np.ndarray, np.ndarray],
    mia_hostname: str,
    mia_secret: str,
    anomed_hostname: str,
    anomed_port: int,
    seed: int | None,
) -> flask.Response:
    """Alias for `membership_inference_attack_validation_data_template`.

    Use this alias."""
    return membership_inference_attack_validation_data_template(
        val, mia_hostname, mia_secret, anomed_hostname, anomed_port, seed
    )


def membership_inference_attack_validation_data_template(
    val: str | Path | tuple[np.ndarray, np.ndarray],
    mia_hostname: str,
    mia_secret: str,
    anomed_hostname: str,
    anomed_port: int,
    seed: int | None,
) -> flask.Response:
    """Respond numpy arrays ("x_val" and "y_val") for training a membership
    inference attack.

    Usually, membership inference attacks need both, data that has been used for
    training of the target PPML model (members) and data that has not been used
    for training of the target PPML model (non-members). This function picks
    random subsets of half the size of `x_val` and `y_val` and responds them, to
    provide "unseen" data for training an MIA.

    Parameters
    ----------
    val : str | Path | tuple[np.ndarray, np.ndarray]
        Either a path to an `.npz` file, containing two arrays with names
        "x_val" and "y_val"; or a tuple of two numpy arrays, namely
        `(x_val, y_val)`. The array named "x_val" contains the features that
        have not been seen by the attack target (non-members). The array named
        "y_val" contains labels/targets that have not been seen by the attack
        victim (non-members).
        Features that have not been seen by the attack target (non-members).
    mia_hostname: str
        The identifying hostname of the membership inference attack.
    mia_secret: str
        The secret obtained at registration of the membership inference attack.
    anomed_hostname: str
        The hostname of the AnoMed platform's main server.
    anomed_port: int
        The port of the AnoMed platform's main server.
    seed : int | None
        The randomness to use for creating the random subsets. If `None`, obtain
        fresh randomness from the environment.

    Returns
    -------
    flask.Response
        A flask response embedding two arrays, with name "x_val" and "y_val",
        having the same datatype and shape (except for the batch dimension) as
        the parameters `x_val` and `y_val`.
    """
    if not _authenticate_hostname(
        mia_hostname, mia_secret, anomed_hostname, anomed_port
    ):
        return _authentication_error(mia_hostname, "MIA")
    (x_val, y_val) = array_utils.load_numpy_array_from_uncertain_source_or_abort(
        val, ("x_val", "y_val")
    )
    return _membership_inference_attack_data(
        x_val, y_val, seed, "val_mia.npz", "x_val", "y_val"
    )


def respond_MIA_attack_success_evaluation_data(
    train: str | Path | tuple[np.ndarray, np.ndarray],
    training_data_seed: int,
    val: str | Path | tuple[np.ndarray, np.ndarray],
    validation_data_seed: int,
    mia_hostname: str,
    mia_secret: str,
    anonymizer_hostname: str,
    anomed_hostname: str,
    anomed_port: int,
    desired_length: int = 100,
    evaluation_data_seed: int | None = None,
) -> flask.Response:
    """Alias for `membership_inference_attack_success_evaluation_data_template`.

    Use this alias."""
    return membership_inference_attack_success_evaluation_data_template(
        train,
        training_data_seed,
        val,
        validation_data_seed,
        mia_hostname,
        mia_secret,
        anonymizer_hostname,
        anomed_hostname,
        anomed_port,
        desired_length,
        evaluation_data_seed,
    )


def membership_inference_attack_success_evaluation_data_template(
    train: str | Path | tuple[np.ndarray, np.ndarray],
    training_data_seed: int,
    val: str | Path | tuple[np.ndarray, np.ndarray],
    validation_data_seed: int,
    mia_hostname: str,
    mia_secret: str,
    anonymizer_hostname: str,
    anomed_hostname: str,
    anomed_port: int,
    desired_length: int = 100,
    evaluation_data_seed: int | None = None,
) -> flask.Response:
    """Generate and respond data to evaluate the success of a membership
    inference attack.

    Usually, MIAs are fitted using member and non-member data, i.e. data that
    was used during the training of the attack target and also data that has not
    been used. To evaluate the success of an MIA, the function draws
    `desired_length` samples (features and labels) from the attack target's
    training set, minus the MIA member training set and `desired_length` samples
    from the attack target's validation set, minus the MIA non-member training
    set. Put differently, pick `desired_length` samples that have been seen by
    the attack target but not by the MIA and `desired_length` samples that have
    not been seen by the attack target and also not by the MIA.

    Parameters
    ----------
    train : str | Path | tuple[np.ndarray, np.ndarray]
        (A path to) a training set (members) of the attack target. If it's a
        path, assume it points to an ".npz` file containing two arrays with
        names "x_train" and "y_train". Otherwise, the pair will be treated as
        `(x_train, y_train)`.
    training_data_seed : int
        The seed used to generate the MIA member data previously (e.g. when
        calling `membership_inference_attack_training_data_template`), to
        recreate the same subset again.
    val : str | Path | tuple[np.ndarray, np.ndarray]
        (A path to) a validation set (non-members) of the attack target. If it's
        a path, assume it points to an ".npz` file containing two arrays with
        names "x_val" and "y_val". Otherwise, the pair will be treated as
        `(x_val, y_val)`.
    validation_data_seed : int
        The seed used to generate the MIA non-member data previously (e.g. when
        calling `membership_inference_attack_validation_data_template`), to
        recreate the same subset again.
    mia_hostname : str
        The identifying hostname of the membership inference attack.
    mia_secret: str
        The secret obtained at registration of the membership inference attack.
    anonymizer_hostname : str
        The identifying hostname of the attack target (anonymizer / PPML model).
    desired_length : int, optional
        The number of sample to draw from members and also from non-members.
        That means the returning array will have length `2 * desired_length`. By
        default 100.
    anomed_hostname: str
        The hostname of the AnoMed platform's main server.
    anomed_port: int
        The port of the AnoMed platform's main server.
    evaluation_data_seed : int | None, optional
        The seed to use when drawing the random subsets of size
        `desired_length`. If `None`, derive a hash value from `mia_hostname` and
        `anonymizer_hostname` to create a (highly likely) unique seed per
        attack-victim combination. By default `None`.

    Returns
    -------
    flask.Response
        Two numpy arrays embedded in a flask response. Their names are "x" and
        "y" and their shapes and datatypes match that of `train` and `val`,
        except for the batch dimension, which is `2 * desired_length`.
    """
    try:
        mia_evaluation_data = membership_inference_attack_success_evaluation_data(
            train,
            training_data_seed,
            val,
            validation_data_seed,
            mia_hostname,
            mia_secret,
            anonymizer_hostname,
            anomed_hostname,
            anomed_port,
            desired_length,
            evaluation_data_seed,
        )
    except (AuthenticationException, AuthorizationException) as e:
        return flask.abort(401, e.args[0])
    return array_utils.respond_numpy_arrays(
        download_name="membership_inference_attack_evaluation_data.npz",
        x=mia_evaluation_data["x"],
        y=mia_evaluation_data["y"],
    )


def membership_inference_attack_success_evaluation_data(
    train: str | Path | tuple[np.ndarray, np.ndarray],
    training_data_seed: int,
    val: str | Path | tuple[np.ndarray, np.ndarray],
    validation_data_seed: int,
    mia_hostname: str,
    mia_secret: str,
    anonymizer_hostname: str,
    anomed_hostname: str,
    anomed_port: int,
    desired_length: int = 100,
    evaluation_data_seed: int | None = None,
) -> dict[str, np.ndarray]:
    """Basically the same as
    `membership_inference_attack_success_evaluation_data_template`, but return
    the evaluation data in a dictionary, instead of embedded in a flask
    response.

    Raises:
    AuthenticationException:
        If `mia_hostname` is not authentic.
    AuthorizationException:
        If `mia_hostname` is not authorized to attack `anonymizer_hostname`.
    """
    if not _authenticate_hostname(
        mia_hostname, mia_secret, anomed_hostname, anomed_port
    ):
        raise AuthenticationException(f"MIA {mia_hostname} is not authentic.")
    if not _authorize_deanonymizer(
        mia_hostname, anonymizer_hostname, anomed_hostname, anomed_port
    ):
        raise AuthorizationException(
            f"MIA {mia_hostname} is not authorized to attack {anonymizer_hostname}."
        )
    (x_train, y_train) = array_utils.load_numpy_array_from_uncertain_source_or_abort(
        train, ["x_train", "y_train"]
    )
    (x_val, y_val) = array_utils.load_numpy_array_from_uncertain_source_or_abort(
        val, ["x_val", "y_val"]
    )

    if evaluation_data_seed is None:
        hash_data = mia_hostname + anonymizer_hostname
        evaluation_data_seed = int(
            hashlib.sha256(hash_data.encode("utf-8")).hexdigest(), 16
        )

    (x_member_unseen, y_member_unseen) = _membership_inference_complementary_data(
        x_train, y_train, training_data_seed
    )
    (x_nonmember_unseen, y_nonmember_unseen) = _membership_inference_complementary_data(
        x_val, y_val, validation_data_seed
    )
    mia_evaluation_data = _draw_mia_evaluation_data(
        x_member_unseen,
        y_member_unseen,
        x_nonmember_unseen,
        y_nonmember_unseen,
        desired_length,
        evaluation_data_seed,
    )
    return mia_evaluation_data


class AuthenticationException(Exception):
    """Raise if hostname authentication fails."""


class AuthorizationException(Exception):
    """Raise if hostname authorization fails."""


def mia_success_eval_data(*args, **kwargs):
    """Short alias for
    `membership_inference_attack_success_evaluation_data_template`."""
    return membership_inference_attack_success_evaluation_data_template(*args, **kwargs)


def _draw_mia_evaluation_data(
    x_member_unseen,
    y_member_unseen,
    x_nonmember_unseen,
    y_nonmember_unseen,
    desired_length,
    seed,
) -> dict[str, np.ndarray]:
    (x_member_unseen, y_member_unseen) = _take_n(
        desired_length,
        x_member_unseen,
        y_member_unseen,
        seed=seed,
    )
    (x_nonmember_unseen, y_nonmember_unseen) = _take_n(
        desired_length,
        x_nonmember_unseen,
        y_nonmember_unseen,
        seed=seed,
    )
    members_mask = np.ones(desired_length, dtype=bool)
    nonmembers_mask = np.zeros(desired_length, dtype=bool)
    x = np.concatenate((x_member_unseen, x_nonmember_unseen))
    y = np.concatenate((y_member_unseen, y_nonmember_unseen))
    memberships = np.concatenate((members_mask, nonmembers_mask))
    [x, y, memberships] = array_utils.shuffles([x, y, memberships], seed=seed)
    return dict(x=x, y=y, memberships=memberships)


def _take_n(
    n: int, arr1: np.ndarray, arr2: np.ndarray, seed: int
) -> tuple[np.ndarray, np.ndarray]:
    [(arr1_l, _), (arr2_l, _)] = array_utils.random_partitions(
        arrays=(arr1, arr2),
        total_length=len(arr1),
        desired_length=n,
        seed=seed,
    )
    return (arr1_l, arr2_l)


def evaluate_MIA_success(
    ground_truth_memberships: np.ndarray,
    estimated_memberships_request: flask.Request,
    anonymizer_hostname: str,
    membership_inference_attack_hostname: str,
    membership_inference_attack_secret: str,
    anomed_hostname: str,
    anomed_port: int,
    challenge_hostname: str,
    challenge_secret: str,
) -> dict[str, float]:
    """Alias for `evaluate_membership_inference_attack_success_template`.

    Use this alias."""
    return evaluate_membership_inference_attack_success_template(
        ground_truth_memberships,
        estimated_memberships_request,
        anonymizer_hostname,
        membership_inference_attack_hostname,
        membership_inference_attack_secret,
        anomed_hostname,
        anomed_port,
        challenge_hostname,
        challenge_secret,
    )


def evaluate_membership_inference_attack_success_template(
    ground_truth_memberships: np.ndarray,
    estimated_memberships_request: flask.Request,
    anonymizer_hostname: str,
    membership_inference_attack_hostname: str,
    membership_inference_attack_secret: str,
    anomed_hostname: str,
    anomed_port: int,
    challenge_hostname: str,
    challenge_secret: str,
) -> dict[str, float]:
    """Calculate the attack success of an attack and submit it to the AnoMed
    platform.

    The success is measured in terms of accuracy, true positive rate and false
    positive rate. Being a member is considered as "positive".

    Parameters
    ----------
    ground_truth_memberships : np.ndarray
        The actual (non)memberships encoded as boolean mask, where `True` means
        "member" and `False` means "non-member".
    estimated_memberships_request : flask.Request
        The estimated (non)memberships, embedded in a flask request object.
    anonymizer_hostname : str
        The identifying hostname of the attack target.
    membership_inference_attack_hostname : str
        The identifying hostname of the MIA.
    membership_inference_attack_secret: str
        The secret obtained at registration of the MIA.
    anomed_hostname : str
        The hostname of the AnoMed platform.
    anomed_port: int
        The port of the AnoMed platform.
    challenge_hostname : str
        The challenge's hostname.
    challenge_secret : str
        The challenge's secret.

    Returns
    -------
    dict[str, float]
        A dictionary containing accuracy ("acc"), true positive rate ("tpr") and
        false positive rate ("fpr").
    """
    if not _authenticate_hostname(
        membership_inference_attack_hostname,
        membership_inference_attack_secret,
        anomed_hostname,
        anomed_port,
    ):
        return _authentication_error(membership_inference_attack_hostname, "MIA")
    if not _authorize_deanonymizer(
        membership_inference_attack_hostname,
        anonymizer_hostname,
        anomed_hostname,
        anomed_port,
    ):
        return flask.abort(
            401,
            f"{membership_inference_attack_hostname} is not authorized to evaluate "
            f"its utility for {anonymizer_hostname}.",
        )
    pred = array_utils.all_numpy_arrays_from_request(estimated_memberships_request)[
        "pred"
    ]
    metrics = _evaluate_membership_inference_attack(
        pred=pred, ground_truth=ground_truth_memberships
    )
    tpr, fpr = metrics["tpr"], metrics["fpr"]

    evaluation = dict(
        anonymizer_hostname=anonymizer_hostname,
        deanonymizer_hostname=membership_inference_attack_hostname,
        tpr=tpr,
        fpr=fpr,
    )
    _submit_single_evaluation(
        anomed_hostname,
        challenge_hostname,
        challenge_secret,
        deanonymizer_evaluation=evaluation,
    )
    return metrics


def eval_mia_success_template(*args, **kwargs) -> dict[str, float]:
    """Short alias for `evaluate_membership_inference_attack_success_template`"""
    return evaluate_membership_inference_attack_success_template(*args, **kwargs)


def _evaluate_membership_inference_attack(
    pred: np.ndarray, ground_truth: np.ndarray
) -> dict[str, float]:
    cm = array_utils.binary_confusion_matrix(pred, ground_truth)
    tp = cm["tp"]
    n = len(pred)
    acc = (tp + cm["tn"]) / n
    tpr = tp / (tp + cm["fn"])
    fpr = cm["fp"] / n
    return dict(acc=acc, tpr=tpr, fpr=fpr)



import os
import sys

sys.path.append(".")

project = "AnoMed Utilities"
copyright = "2025, Yannik Potdevin"
author = "Yannik Potdevin"

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.napoleon",  # to parse NumPy docstrings
    "myst_parser",  # to parse .md files too
    "autodoc2",  # this is more compatible with MyST than Sphinx' autodoc
]

templates_path = ["_templates"]
html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "/")
html_theme = "sphinx_rtd_theme"

myst_enable_extensions = [
    "fieldlist",
]

autodoc2_packages = [
    "../src/anomed_utils",
]
autodoc2_docstring_parser_regexes = [
    # this will render all docstrings as Markdown
    (r".*", "docstrings_parser"),
]
autodoc2_module_all_regexes = [
    r"anomed_utils\..*",
]
autodoc2_hidden_objects = ["private", "inherited"]



from pathlib import Path
from typing import Any
from unittest.mock import MagicMock

import falcon
import falcon.app
import numpy as np
import pandas as pd
import pytest
import requests
from falcon import testing

from anomed_utils import web


@pytest.fixture()
def empty_array():
    return np.array([])


@pytest.fixture()
def int_array():
    return np.arange(10)


@pytest.fixture()
def float_array():
    return np.arange(10) + 0.5


@pytest.fixture()
def object_array():
    return np.array(3 * ["foo", "bar", "baz"])


class Dummy:
    def fit(self, X: np.ndarray, y: np.ndarray) -> None:
        pass

    def save(self, filepath: str | Path) -> None:
        with open(filepath, "w") as file:
            file.write("test")

    def validate_input(self, X: np.ndarray) -> None:
        if X.dtype != np.int_:
            raise ValueError()


@pytest.fixture()
def client(tmp_path, int_array):
    app = falcon.App()
    app.add_route("/", web.StaticJSONResource(dict(message="hello world")))
    app.add_route(
        "/fit",
        web.FitResource(
            data_getter=lambda: dict(X=int_array, y=int_array),
            model=Dummy(),
            model_filepath=tmp_path / "model",
        ),
    )
    return testing.TestClient(app=app)


def test_StaticJSONResource(client):
    json = dict(message="hello world")
    response = client.simulate_get("/")
    assert response.json == json


def test_FitResource(client):
    response = client.simulate_post("/fit")
    assert response.status == falcon.HTTP_CREATED
    assert response.json == dict(message="Fitting has been completed successfully.")


def test_array_to_bytes_conversion(empty_array, int_array, float_array, object_array):
    expected_result = dict(
        empty=empty_array, ints=int_array, floats=float_array, objs=object_array
    )
    result = web.bytes_to_named_ndarrays(web.named_ndarrays_to_bytes(expected_result))
    _assert_dict_array_eq(expected_result, result)


def _assert_dict_array_eq(
    expected_result: dict[Any, np.ndarray], result: dict[Any, np.ndarray]
) -> None:
    expected_keys = expected_result.keys()
    assert expected_keys == result.keys()
    for key in expected_keys:
        assert np.array_equal(expected_result[key], result[key])


def test_bytes_to_named_ndarrays_or_raise(int_array):
    arrays = dict(a=int_array, b=int_array)
    payload = web.named_ndarrays_to_bytes(arrays)
    _assert_dict_array_eq(
        arrays, web.bytes_to_named_ndarrays_or_raise(payload, ["a", "b"], 500)
    )


def test_failing_bytes_to_named_ndarrays_or_raise(int_array):
    arrays = dict(a=int_array, b=int_array)
    payload = web.named_ndarrays_to_bytes(arrays)
    with pytest.raises(falcon.HTTPError):
        _assert_dict_array_eq(
            arrays, web.bytes_to_named_ndarrays_or_raise(payload, ["a", "c"], 500)
        )


def test_get_named_arrays_or_raise(int_array, mocker):
    arrays = dict(a=int_array, b=int_array)
    mock = _mock_get_numpy_arrays(mocker, named_arrays=arrays)
    _assert_dict_array_eq(
        web.get_named_arrays_or_raise("example.com", ["a", "b"]), arrays
    )
    mock.assert_called_once()


def test_failing_get_named_arrays_or_raise(int_array, mocker):
    mock = _mock_get_connection_error(mocker)
    with pytest.raises(falcon.HTTPServiceUnavailable):
        web.get_named_arrays_or_raise("example.com", ["a", "b"])
    mock.assert_called_once()


def _mock_get_numpy_arrays(
    _mocker, named_arrays: dict[str, np.ndarray], status_code: int = 200
) -> MagicMock:
    mock_response = _mocker.MagicMock()
    mock_response.status_code = status_code
    mock_response.content = web.named_ndarrays_to_bytes(named_arrays)
    return _mocker.patch("requests.get", return_value=mock_response)


def _mock_get_connection_error(_mocker) -> MagicMock:
    mock_response = _mocker.MagicMock()
    mock_response.side_effect = requests.ConnectionError()
    return _mocker.patch("requests.get", return_value=mock_response)


def test_dataframe_bytes_conversion():
    orig_dfs = []
    orig_dfs.append(
        pd.DataFrame(
            {
                "a": list("abc"),
                "b": list(range(1, 4)),
                "c": np.arange(3, 6).astype("u1"),
                "d": np.arange(4.0, 7.0, dtype="float64"),
                "e": [True, False, True],
                "f": pd.Categorical(list("abc")),
                "g": pd.date_range("20130101", periods=3),
                "h": pd.date_range("20130101", periods=3, tz="US/Eastern"),
                "i": pd.date_range("20130101", periods=3, freq="ns"),
            }
        )
    )
    orig_dfs.append(pd.DataFrame(data=[1, 2], index=[True, False]))
    orig_dfs.append(pd.DataFrame(data=[1, 2, 3], index=["a", "b", "c"]))
    orig_dfs.append(
        pd.DataFrame(data=[1, 2, 3], index=pd.date_range("20130101", periods=3))
    )
    for orig_df in orig_dfs:
        bytes = web.dataframe_to_bytes(orig_df)
        restored_df = web.bytes_to_dataframe(bytes)
        assert orig_df.equals(restored_df) and np.array_equal(
            orig_df.columns, restored_df.columns
        )



from .arrays import (
    binary_confusion_matrix,
    random_partitions,
    shuffles,
)
from .web import (
    FitResource,
    StaticJSONResource,
    bytes_to_dataframe,
    bytes_to_named_ndarrays,
    bytes_to_named_ndarrays_or_raise,
    dataframe_to_bytes,
    get_named_arrays_or_raise,
    named_ndarrays_to_bytes,
)

__all__ = [
    "bytes_to_dataframe",
    "bytes_to_named_ndarrays",
    "bytes_to_named_ndarrays_or_raise",
    "binary_confusion_matrix",
    "dataframe_to_bytes",
    "FitResource",
    "get_named_arrays_or_raise",
    "named_ndarrays_to_bytes",
    "random_partitions",
    "shuffles",
    "StaticJSONResource",
]



from pathlib import Path
from typing import Any, Callable

import flask
import numpy as np

from . import arrays as array_utils
from . import web as web_utils


def main_page_template() -> flask.Response:
    return flask.make_response("Server is running.")


def initiate_supervised_anonymizer_training_template(
    challenge_hostname: str,
    train_model: Callable[[np.ndarray, np.ndarray], Any],
    save_model: Callable[[Any], None],
) -> flask.Response:
    """Initiate the training of an anonymizer/PPML model in the setting of
    supervised learning.

    Parameters
    ----------
    challenge_hostname : str
        The identifying hostname of the challenge to obtain the training data
        from.
    train_model : Callable[[np.ndarray, np.ndarray], Any]
        A function which, when receiving training data (first argument being
        features, second argument being labels/targets), returns an anonymizer
        trained on that data. The returned anonymizer should be compatible with
        `save_model`.
    save_model : Callable[[Any], None]
        A function which, when applied to the result of `train_model`, saves
        that model to disk.

    Returns
    -------
    flask.Response
        A simple message indicating that training has been completed.
    """
    training_data = array_utils.numpy_arrays_from_remote(
        host=challenge_hostname, route="/data/anonymizer/training"
    )
    model = train_model(training_data["x_train"], training_data["y_train"])
    save_model(model)
    return flask.make_response("Training completed")


def initiate_supervised_anonymizer_validation_template(
    model_path: str | Path,
    load_model: Callable[[Path], Any],
    challenge_hostname: str,
    anonymizer_hostname: str,
    anonymizer_secret: str,
    predict: Callable[[np.ndarray, Any], np.ndarray],
) -> flask.Response:
    """Initiate the validation of an anonymizer in the setting of supervised
    learning.

    Parameters
    ----------
    model_path : str | Path
        Where to load the model to evaluate from.
    load_model : Callable[[Path], Any]
        The function to actually load the function. The ambient function will
        take care of the case in which the `model_path` does not exist.
    challenge_hostname : str
        The identifying hostname of the challenge to draw validation data from
        and to use the utility evaluation functionality from.
    anonymizer_hostname : str
        The identifying hostname of the anonymizer being validated.
    predict : Callable[[np.ndarray, Any], np.ndarray]
        A function taking an array of features and the loaded model (result of
        `load_model`) and returning an array of labels/targets.

    Returns
    -------
    flask.Response
        A message indicating the completion of the validation and containing the
        response of the challenge (usually some evaluation metrics).
    """
    model_path = Path(model_path)
    model = web_utils.apply_or_abort(
        condition=model_path.exists(),
        error_msg="No model to load on disk. Still need to train?",
    )(load_model)(model_path)
    validation_data = array_utils.numpy_arrays_from_remote(
        host=challenge_hostname,
        route=f"/data/anonymizer/validation/{anonymizer_hostname}",
        credentials=dict(anonymizer_secret=anonymizer_secret),
    )
    pred = predict(validation_data["x_val"], model)
    response = array_utils.numpy_arrays_to_remote(
        host=challenge_hostname,
        route=f"/evaluation/utility/anonymizer/{anonymizer_hostname}",
        named_arrays=dict(pred=pred),
    )
    response.raise_for_status()
    return flask.make_response(f"Validation completed. Response: {response.text}")


def predict_supervised_anonymizer_template(
    model_path: str | Path,
    load_model: Callable[[Path], Any],
    request: flask.Request,
    predict_batched: Callable[[np.ndarray, Any, int | None], np.ndarray],
) -> flask.Response:
    """Load an anonymizer/PPML model and use it to predict the labels/targets of
    a NumPy array of features.

    Parameters
    ----------
    model_path : str | Path
        Where to load the estimator from.
    load_model : Callable[[Path], Any]
        The function to actually load the estimator (result should be compatible
        with second argument of `predict_batched`).
    request : flask.Request
        The request object carrying the NumPy array of features named "x".
    predict_batched : Callable[[np.ndarray, Any, int  |  None], np.ndarray]
        A function taking a feature vector (first argument), an estimator as
        returned by `load_model` and optionally a batch size to estimate the
        labels/targets of that feature vector.

    Returns
    -------
    flask.Response
        A NumPy array with name "pred" embedded in a flask response object,
        containing the model's estimate.
    """
    model_path = Path(model_path)
    model = web_utils.apply_or_abort(
        condition=model_path.exists(),
        error_msg="No model to load on disk. Still need to train?",
    )(load_model)(model_path)
    (x,) = array_utils.numpy_arrays_from_request(request, "x")
    try:
        batch_size = int(request.args["batch_size"])
    except (KeyError, ValueError):
        batch_size = None
    prediction = predict_batched(x, model, batch_size)
    return array_utils.respond_numpy_arrays(pred=prediction)


def compute_loss_supervised_anonymizer_template(
    model_path: str | Path,
    load_model: Callable[[Path], Any],
    request: flask.Request,
    compute_loss_batched: Callable[
        [np.ndarray, np.ndarray, Any, int | None], np.ndarray
    ],
) -> flask.Response:
    """Load an anonymizer/PPML model and use it to calculate the element-wise
    loss of its prediction compared to the ground truth.

    Parameters
    ----------
    model_path : str | Path
        Where to load the estimator from.
    load_model : Callable[[Path], Any]
        The function to actually load the estimator (result should be compatible
        with third argument of `compute_loss`).
    request : flask.Request
        The request object carrying the NumPy arrays of features (named "x") and
        ground truth (named "y").
    compute_loss_batched : Callable[
            [np.ndarray, np.ndarray, Any, int  |  None], np.ndarray
        ]
        A function taking a feature vector (first argument), the ground truth
        (second argument), an estimator (third argument) as returned by
        `load_model` and optionally a batch size to calculate the element-wise
        loss of the model's prediction compared to the ground truth.

    Returns
    -------
    flask.Response
        A NumPy array with name "loss" embedded in a flask response object,
        containing the element-wise loss.
    """
    model_path = Path(model_path)
    model = web_utils.apply_or_abort(
        condition=model_path.exists(),
        error_msg="No model to load on disk. Still need to train?",
    )(load_model)(model_path)
    (x, y) = array_utils.numpy_arrays_from_request(request, "x", "y")
    try:
        batch_size = int(request.args["batch_size"])
    except (KeyError, ValueError):
        batch_size = None
    loss = compute_loss_batched(x, y, model, batch_size)
    return array_utils.respond_numpy_arrays(loss=loss)



# From https://github.com/sphinx-extensions2/sphinx-autodoc2/issues/33#issuecomment-2664238481
from docutils import nodes
from myst_parser.parsers.sphinx_ import MystParser
from sphinx.ext.napoleon import docstring


class NapoleonParser(MystParser):
    def parse(self, input_string: str, document: nodes.document) -> None:
        parsed_content = str(
            docstring.GoogleDocstring(str(docstring.NumpyDocstring(input_string)))
        )
        return super().parse(parsed_content, document)


Parser = NapoleonParser



from pathlib import Path
from typing import Any, Callable

import flask
import numpy as np
import utils.arrays as array_utils
import utils.web as web_utils

from art.estimators import BaseEstimator
from art.estimators.classification import ClassifierMixin
from art.estimators.regression import RegressorMixin


class WebClassifier(ClassifierMixin, BaseEstimator):
    """A wrapper class to make anonymizers (classifier), which are only
    accessible via web, available to ART's membership inference attack.
    """

    def __init__(
        self,
        url: str,
        input_shape: tuple[int, ...],
        nb_classes: int,
        clip_values: tuple[float, float] | None = None,
        x_name: str = "x",
        pred_name: str = "pred",
    ):
        """Instantiate a WebClassifier wrapper.

        Parameters
        ----------
        url : str
            The url to a POST API accepting a numpy array with the name
            `x_name`.
        input_shape : Tuple[int, ...]
            The shape of the array (except for the batch dimension) the API at
            `url` expects.
        nb_classes : int
            The number of classes the classifier at `url` is able to
            distinguish.
        x_name : str
            The name of the numpy array the API at `url` expects as input. By
            default "x".
        pred_name : str
            The name of the numpy array the API at `url` returns. By default
            "pred".
        """
        self.url = url
        self._input_shape = input_shape
        self._nb_classes = nb_classes
        self.x_name = x_name
        self.pred_name = pred_name
        self._clip_values = clip_values

    def predict(self, x: np.ndarray, batch_size: int = 128) -> np.ndarray:
        """Feed an array via web to an anonymizer and retrieve the prediction
        back from the response.

        Parameters_description_, by default "x"
            The features.
        batch_size : int
            The batch size to use for prediction, by default 128.

        Returns
        -------
        prediction: np.ndarray
            The prediction.
        """
        arrays = {self.x_name: x}
        query_params = dict(batch_size=batch_size)
        response = array_utils.post_numpy_arrays(self.url, arrays, query_params)
        response.raise_for_status()
        (array,) = array_utils.numpy_arrays_from_response(response, self.pred_name)
        return array

    def fit(self):
        """This has to be overridden because of the base class'
        `@abstractmethod` decorator - however it actually is not needed for the
        membership inference attack. Just ignore this function.

        Raises
        ------
        NotImplementedError
            Immediately.
        """
        raise NotImplementedError

    @property
    def input_shape(self) -> tuple[int, ...]:
        return self._input_shape

    @property
    def nb_classes(self) -> int:
        return self._nb_classes

    @property
    def clip_values(self) -> tuple[float, float] | None:
        return self._clip_values


class WebRegressor(RegressorMixin, BaseEstimator):
    """A wrapper class to make anonymizers (regressor), which are only
    accessible via web, available to ART's membership inference attack.
    """

    def __init__(
        self,
        url: str,
        input_shape: tuple[int, ...],
        x_name: str = "x",
        y_name: str = "y",
        losses_name: str = "loss",
    ):
        """Instantiate a WebRegressor wrapper.

        Parameters
        ----------
        url : str
            The url to a POST API accepting numpy arrays with the names
            `x_name` and `y_name`.
        input_shape : tuple[int, ...]
            The shape of the array (except for the batch dimension) the API at
            `url` expects for the feature array.
        x_name : str, optional
            The name of the numpy array the API at `url` expects as feature
            input. By default "x".
        y_name : str, optional
            The name of the numpy array the API at `url` expects as target
            input. By default "y".
        losses_name : str, optional
            The name of the numpy array the API at `url` returns. By default
            "loss".
        """
        self.url = url
        self._input_shape = input_shape
        self.x_name = x_name
        self.y_name = y_name
        self.losses_name = losses_name

    def compute_loss(
        self, x: np.ndarray, y: np.ndarray, batch_size: int = 128
    ) -> np.ndarray:
        """Feed arrays via web to an anonymizer and retrieve the loss vector
        back from the response.

        Parameters
        ----------
        x : np.ndarray
            The NumPy array containing the features, as expected by the
            regressor.
        y : np.ndarray
            The NumPy array containing the targets, as expected by the
            regressor.
        batch_size : int, optional
            The batch size to use during loss calculation (to avoid system
            overload), by default 128.

        Returns
        -------
        np.ndarray
            The NumPy array containing the element-wise loss values, as returned
            by the regressor.
        """
        arrays = {self.x_name: x, self.y_name: y}
        query_params = dict(batch_size=batch_size)
        response = array_utils.post_numpy_arrays(self.url, arrays, query_params)
        response.raise_for_status()
        (array,) = array_utils.numpy_arrays_from_response(response, self.losses_name)
        return array

    def predict(self):
        """This has to be overridden because of the base class'
        `@abstractmethod` decorator - however it actually is not needed for the
        membership inference attack. Just ignore this function.

        Raises
        ------
        NotImplementedError
            Immediately.
        """
        raise NotImplementedError

    def fit(self):
        """This has to be overridden because of the base class'
        `@abstractmethod` decorator - however it actually is not needed for the
        membership inference attack. Just ignore this function.

        Raises
        ------
        NotImplementedError
            Immediately.
        """
        raise NotImplementedError

    @property
    def input_shape(self) -> tuple[int, ...]:
        return self._input_shape


def main_page_template() -> flask.Response:
    return flask.make_response("Server is running.")


def init_MIA_fitting(
    target: Any,
    members: (
        str
        | Path
        | tuple[str, str]
        | tuple[np.ndarray, np.ndarray]
        | dict[str, np.ndarray]
    ),
    nonmembers: (
        str
        | Path
        | tuple[str, str]
        | tuple[np.ndarray, np.ndarray]
        | dict[str, np.ndarray]
    ),
    credentials: dict[str, str] | None,
    fit: Callable[[Any, np.ndarray, np.ndarray, np.ndarray, np.ndarray], Any],
    save: Callable[[Any], None],
) -> flask.Response:
    """Alias for `initiate_membership_inference_attack_fitting_template`.

    Use this alias."""
    return initiate_membership_inference_attack_fitting_template(
        target, members, nonmembers, credentials, fit, save
    )


def initiate_membership_inference_attack_fitting_template(
    target: Any,
    members: (
        str
        | Path
        | tuple[str, str]
        | tuple[np.ndarray, np.ndarray]
        | dict[str, np.ndarray]
    ),
    nonmembers: (
        str
        | Path
        | tuple[str, str]
        | tuple[np.ndarray, np.ndarray]
        | dict[str, np.ndarray]
    ),
    credentials: dict[str, str] | None,
    fit: Callable[[Any, np.ndarray, np.ndarray, np.ndarray, np.ndarray], Any],
    save: Callable[[Any], None],
) -> flask.Response:
    """Initiate training/fitting of a membersihp inference attack and persist
    it.

    Parameters
    ----------
    target : Any
        An attack target, compatible with `fit` (it will be the first argument
        given to `fit`).
    members : str
              | Path
              | tuple[str, str]
              | tuple[np.ndarray, np.ndarray]
              | dict[str, np.ndarray]
        Data that has been used during the training of `target` (i.e. was part
        of target's training data). It may be a path to an ".npz" file
        containing two arrays with the names "x_train" and "y_train"; it may be
        a pair of hostname and route (`(str, str)`), which will respond NumPy
        arrays with names "x_train" and "y_train"; it may be a pair of feature
        and label NumPy arrays; or a dictionary of NumPy arrays with keys
        "x_train" and "y_train".
    nonmembers : str
                 |  Path
                 |  tuple[str, str]
                 |  tuple[np.ndarray, np.ndarray]
                 |  dict[str, np.ndarray]
        Data that has not been used during the training of `target` (i.e. was
        not part of target's training data, but, for example, part of its
        validation data). It may be a path to an ".npz" file containing two
        arrays with the names "x_train" and "y_train"; it may be a pair of
        hostname and route (`(str, str)`), which will respond NumPy arrays with
        names "x_train" and "y_train"; it may be a pair of feature and label
        NumPy arrays; or a dictionary of NumPy arrays with keys "x_train" and
        "y_train".
    credentials : dict[str, str] | None
        If `nonmembers` consists of hostname and route, this might be necessary.
    fit : Callable[[Any, np.ndarray, np.ndarray, np.ndarray, np.ndarray], Any]
        A function that takes an attack target (first argument), member
        features, member labels, non-member features and non-member labels and
        returns a fitted membership inference attack (compatible with `save`).
    save : Callable[[Any], None]
        A function that takes a membership inference attack as returned by `fit`
        and persists it, by saving it to disk.

    Returns
    -------
    flask.Response
        A simple message indicating the success of fitting the MIA.
    """
    (x_train, y_train) = array_utils.load_numpy_array_from_uncertain_source_or_abort(
        members, ("x_train", "y_train")
    )
    (x_val, y_val) = array_utils.load_numpy_array_from_uncertain_source_or_abort(
        nonmembers, ("x_val", "y_val"), credentials=credentials
    )
    atk = fit(target, x_train, y_train, x_val, y_val)
    save(atk)
    return flask.make_response("Attack fitted")


def init_MIA_success_evaluation(
    mia_path: str | Path,
    load_mia: Callable[[Path], Any],
    estimate_memberships: Callable[[np.ndarray, np.ndarray, Any], np.ndarray],
    challenge_hostname: str,
    mia_hostname: str,
    mia_secret: str,
    victim_hostname: str,
) -> flask.Response:
    """Alias for `initiate_membership_inference_attack_success_evaluation_template`.

    Use this alias."""
    return initiate_membership_inference_attack_success_evaluation_template(
        mia_path,
        load_mia,
        estimate_memberships,
        challenge_hostname,
        mia_hostname,
        mia_secret,
        victim_hostname,
    )


def initiate_membership_inference_attack_success_evaluation_template(
    mia_path: str | Path,
    load_mia: Callable[[Path], Any],
    estimate_memberships: Callable[[np.ndarray, np.ndarray, Any], np.ndarray],
    challenge_hostname: str,
    mia_hostname: str,
    mia_secret: str,
    victim_hostname: str,
) -> flask.Response:
    """Load a membership inference attack, obtain evaluation data, let the
    attack estimate memberships and submit the estimation to the AnoMed platform
    where they will be evaluated.

    Parameters
    ----------
    mia_path : str | Path
        Where to obtain the fitted membership inference attack from.
    load_mia : Callable[[Path], Any]
        A function actually loading the MIA from disk (loaded MIA should be
        compatible with `estimate_memberships`). The ambient function will take
        care of the case, if `mia_path` does not exist.
    estimate_memberships : Callable[[np.ndarray, np.ndarray, Any], np.ndarray]
        Given samples (first argument being features, second argument being
        labels) and a membership inference attack (third argument) as returned
        by `load_mia`, this function estimates the membership status of the
        samples. The returned array is expected to be boolean and
        `load_mia(features, labels, mia)[i] == True` if and only if
        `(features[i], labels[i])` is estimated to be a member.
    challenge_hostname : str
        The identifying hostname of the challenge where to obtain evaluation
        data from and where to submit the estimations to.
    mia_hostname : str
        The identifying hostname of the membership inference attack being
        evaluated.
    mia_secret : str
        The secret obtained at registration of the MIA.
    victim_hostname : str
        The identifying hostname of the attack target.

    Returns
    -------
    flask.Response
        A simple message indicating the success of evaluating the MIA.
    """
    mia_path = Path(mia_path)
    mia = web_utils.apply_or_abort(
        mia_path.exists(),
        error_msg="No fitted attack found. Did you already train one?",
    )(load_mia)(mia_path)
    eval_data = array_utils.numpy_arrays_from_remote(
        host=challenge_hostname,
        route=(
            "/data/deanonymizer/attack-success-evaluation/"
            f"{mia_hostname}/{victim_hostname}"
        ),
        credentials=dict(deanonymizer_secret=mia_secret),
    )
    x, y = eval_data["x"], eval_data["y"]
    estimated_memberships = estimate_memberships(x, y, mia)
    resp = array_utils.numpy_arrays_to_remote(
        host=challenge_hostname,
        route=("/evaluation/utility/deanonymizer/" f"{mia_hostname}/{victim_hostname}"),
        named_arrays=dict(pred=estimated_memberships),
    )
    resp.raise_for_status()
    return flask.make_response(f"Evaluation completed. Response: {resp.text}")
